/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Boichat                                 *
 *   nicolas@boichat.ch                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 *   DDC is a Trademark of VESA (Video Electronics Standard Association).  *
 *                                                                         *
 ***************************************************************************/
/*
 * EDID Parser (Linux kernel) :
 * 
 * Copyright (C) 2002 James Simmons <jsimmons@users.sf.net>
 *
 * The EDID Parser is a conglomeration from the following sources:
 *
 *   1. SciTech SNAP Graphics Architecture
 *      Copyright (C) 1991-2002 SciTech Software, Inc. All rights reserved.
 *
 *   2. XFree86 4.3.0, interpret_edid.c
 *      Copyright 1998 by Egbert Eich <Egbert.Eich@Physik.TU-Darmstadt.DE>
 * 
 *   3. John Fremlin <vii@users.sourceforge.net> and 
 *      Ani Joshi <ajoshi@unixbox.com>
 */

#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/i2c.h>
#include <linux/i2c-dev.h>
#include <dirent.h>

#include "ddccontrol.h"

#include "edid.h"

#define PAUSE 30000 /* Pause between packets */

struct monitor {
	char* manufacturer;
	int model;
	char* name;
};

struct monitor M(char* manufacturer, int model, char* name) {
	struct monitor m;
	m.manufacturer = manufacturer;
	m.model = model;
	m.name = name;
	return m;
}

const char* manufacturers[][2] = {
	{"MEL", "Mitsubishi"}, 
	{NULL, ""}};

const struct monitor models[] = {
	M("MEL", 0x4513, "Diamond Pro 2060u"),
	M(NULL, -1, "")
};

ddc_control::ddc_control(int bus) {
	this->bus = bus;
	fd = -1;
	error = DDC_NO_ERROR;
	manufacturer[0] = 0;
	model = 0;
	serial = 0;
	year = 0;
	week = 0;
	name[0] = 0;
}

ddc_control::~ddc_control() {
	if (fd != -1) {
		close(fd);
	}
}

int ddc_control::init() {
	char device[32];
	snprintf(device, 32, "/dev/i2c/%d", bus);

	fd = open(device, O_RDWR);
	if (fd == -1) {
		printf("Error, can't open device (%s).\n", strerror(errno));
		if (errno == EACCES) {
			error = DDC_PERMISSION_DENIED;
		}
		else if (errno == ENOENT) {
			error = DDC_NO_SUCH_FILE_ERROR;
		}
		else {
			error = DDC_UNKNOWN_OPEN_ERROR;
		}
		fd = -1;
		return 0;
	}

	return get_EDID();
}

int ddc_control::get_EDID() {
	if (fd == -1) {
		error = DDC_NOT_INITED;
		return 0;
	}

	int cnt;
	unsigned char buffer[128];

	int addr = 0x50; //DDC2
	if (ioctl(fd,I2C_SLAVE,addr) < 0) {
		printf("Error, can't set slave device (%s).\n", strerror(errno));
		return 0;
	}

	buffer[0] = 0x00;

	cnt = write(fd, buffer, 1);
//	printf("Written: %d\n", cnt);

	if (cnt != 1) {
		printf("Could not write all bytes.\n");
		error = DDC_WRITE_ERROR;
		return 0;
	}

	cnt = read(fd, buffer, 128);
//	printf("Read: %d\n", cnt);

	if (cnt != 128) {
		printf("Could not read all bytes.\n");
		error = DDC_READ_ERROR;
		return 0;
	}

/*	for (int i = 0; i < 16; i++) {
		printf("%#x ", buffer[i] & 0xFF);
	}
	printf("\n");*/

	if ((!edid_checksum(buffer)) || (!edid_check_header(buffer))) {
		return DDC_INVALID_EDID;
	}

	manufacturer[0] = ((buffer[ID_MANUFACTURER_NAME] & 0x7c) >> 2) + '@';
	manufacturer[1] = ((buffer[ID_MANUFACTURER_NAME] & 0x03) << 3) + 
	                  ((buffer[ID_MANUFACTURER_NAME+1] & 0xe0) >> 5) + '@';
	manufacturer[2] = (buffer[ID_MANUFACTURER_NAME+1] & 0x1f) + '@';
	manufacturer[3] = 0;
	model = buffer[ID_MODEL] + (buffer[ID_MODEL+1] << 8);
	serial = buffer[ID_SERIAL_NUMBER] + (buffer[ID_SERIAL_NUMBER+1] << 8) +
	       (buffer[ID_SERIAL_NUMBER+2] << 16) + (buffer[ID_SERIAL_NUMBER+3] << 24);
	year = buffer[MANUFACTURE_YEAR] + 1990;
	week = buffer[MANUFACTURE_WEEK];

	error = DDC_NO_ERROR;
	return 1;
}

int ddc_control::getValue(unsigned char index, unsigned char* value, unsigned char* max) {
	if (fd == -1) {
		error = DDC_NOT_INITED;
		return 0;
	}

	int cnt;
	unsigned char buffer[128];

	int addr = 0x37; //DDC/CI
	if (ioctl(fd,I2C_SLAVE,addr) < 0) {
		printf("Error, can't set slave device (%s).\n", strerror(errno));
		return 1;
	}

	buffer[0] = 0x51;
	buffer[1] = 0x82;
	buffer[2] = 0x01;
	buffer[3] = index;
	addchecksum(buffer);

	usleep(PAUSE);

	cnt = write(fd, buffer, 5);
//	printf("Written: %d\n", cnt);

	if (cnt != 5) {
		printf("Could not write all bytes.\n");
		error = DDC_WRITE_ERROR;
		return 0;
	}

	usleep(PAUSE);

	cnt = read(fd, buffer, 11);
//	printf("Read: %d\n", cnt);

	if (cnt != 11) {
		printf("Could not read all bytes.\n");
		error = DDC_READ_ERROR;
		return 0;
	}

/*	for (int i = 0; i < cnt; i++) {
		printf("%#x ", buffer[i] & 0xFF);
	}*/

	if (!checksum(buffer, 0x51)) {
		printf("Bad read checksum...\n");
		error = DDC_READCHKSUM_ERROR;
		return 0;
	}

	if (	(buffer[1] != 0x88) || (buffer[2] != 0x02) || 
			(buffer[3] != 0x00) || (buffer[4] != index) ||
			(buffer[5] != 0x00) || (buffer[6] != 0x00) || 
			(buffer[8] != 0x00)) {
		printf("Bad answer...\n");
		error = DDC_BAD_ANSWER;
		return 0;		
	}

	*max   = buffer[7];
	*value = buffer[9];

	error = DDC_NO_ERROR;
	return 1;
}

int ddc_control::setValue(unsigned char index, unsigned char value) {
	if (fd == -1) {
		error = DDC_NOT_INITED;
		return 0;
	}

	int cnt;
	unsigned char buffer[128];

	usleep(PAUSE);

	buffer[0] = 0x51;
	buffer[1] = 0x84;
	buffer[2] = 0x03;
	buffer[3] = index;
	buffer[4] = 0x00;
	buffer[5] = value;
	addchecksum(buffer);

	cnt = write(fd, buffer, 7);
//	printf("Written: %d\n", cnt);	

	if (cnt != 7) {
		printf("Could not write all bytes.\n");
		error = DDC_WRITE_ERROR;
		return 0;
	}
	
	error = DDC_NO_ERROR;
	return 1;
}

/* add checksum to a write packet */
void ddc_control::addchecksum(unsigned char* buffer) {
	char chksum = 0x6E; /* first byte, [I2C address << 1] */
	int len = (buffer[1] & 0x7F) + 2;
	for (int i = 0; i < len; i++) {
		chksum ^= buffer[i];
	}
	buffer[len] = chksum;
}

/* Check sum of a read packet */
int ddc_control::checksum(unsigned char* buffer, unsigned char subaddr) {
	char chksum = 0x6F; /* first byte, [(I2C address << 1) | 0x01] */
	if (buffer[0] != 0x6E)
		return 0;

	chksum ^= subaddr; /* See specs */

	int len = (buffer[1] & 0x7F) + 3;
	for (int i = 1; i < len; i++) {
		chksum ^= buffer[i];
	}

	return (chksum == 0);
}

int ddc_control::edid_checksum(unsigned char *edid) {
	unsigned char i, csum = 0;

	for (i = 0; i < EDID_LENGTH; i++)
		csum += edid[i];

	if (csum == 0x00) {
		/* checksum passed, everything's good */
		return 1;
	} else {
		printf("EDID checksum failed, aborting\n");
		return 0;
	}
}

int ddc_control::edid_check_header(unsigned char *edid) {
	if ((edid[0] != 0x00) || (edid[1] != 0xff) || (edid[2] != 0xff) ||
	    (edid[3] != 0xff) || (edid[4] != 0xff) || (edid[5] != 0xff) ||
	    (edid[6] != 0xff)) {
		printf("EDID header doesn't match EDID v1 header, aborting\n");
		return 0;
	}
	return 1;
}

int ddc_control::getError() {
	return error;
}

int ddc_control::getBus() {
	return bus;
}

const char* ddc_control::getManufacturer() {
	return (const char*)&manufacturer;
}

const int ddc_control::getModel() {
	return model;
}

const long ddc_control::getSerial() {
	return serial;
}

const int ddc_control::getYear() {
	return year;
}

const int ddc_control::getWeek() {
	return week;
}

const char* ddc_control::getName() {
	if (name[0] != NULL)
		return (const char*)&name;

	int manindex;
	int modindex;
	
	for (manindex = 0;; manindex++) {
		if (manufacturers[manindex][0] == NULL) {
			manindex = -1;
			break;
		}
		else if (strcmp(manufacturers[manindex][0], manufacturer) == 0) {
			break;
		}
	}
	
	if (manindex == -1) {
		modindex = -1;
	}
	else {
		for (modindex = 0;; modindex++) {
			if (models[modindex].manufacturer == NULL) {
				modindex = -1;
				break;
			}
			else if ((strcmp(models[modindex].manufacturer, manufacturer) == 0) && // Manufacturer matches
				(models[modindex].model == model)) {
				break;
			}
		}
	}
	
	if (manindex == -1) {
		snprintf(name, 64, "Unknown manufacturer (%s) and model (%x)",
			manufacturer, model);
	}
	else if (modindex == -1) {
		snprintf(name, 64, "%s, unknown model (%x)", manufacturers[manindex][1], model);
	}
	else {
		snprintf(name, 64, "%s %s", manufacturers[manindex][1], models[modindex].name);
	}
		
	return (const char*)&name;
}
